#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "hash.h"

static const size_t BUFLEN = 32;
static const uint32_t MAXINSERTS = 100000;

/* Matches the hash_hasher definition in hash.h: */
static uint64_t hash_fn(const void* k) {
	uint64_t hash_val = 0;
	uint64_t coefficient = 1;

	for (const char* p = (const char*) k; *p != '\0'; p++) {
		hash_val += coefficient * (*p);
		coefficient *= 37;
	}

	return hash_val;
}

/* Matches the hash_compare definition in hash.h. This function compares
 * two keys that are strings. */
static int hash_strcmp(const void* k1, const void* k2) {
	return strcmp((const char*) k1, (const char*) k2);
}

int main(int argc, char* argv[]) {
	int N;
	char* not_found_key = "not-found key";
	hash_table* ht;
	char* k;
	long int* v;
	char* removed_key;
	long int* removed_value;
	char strbuf[BUFLEN];

	/* Check for correct invocation: */
	if (argc != 2) {
		printf("Usage: %s <N>\n"
				"Run test inserting a total of N items\n", argv[0]);
		exit(1);
	}
	N = atoi(argv[1]);
	if (N <= 0 || N > MAXINSERTS) {
		N = MAXINSERTS;
	}

	/* Create the hash table. */
	ht = hash_create(hash_fn, hash_strcmp);
	if (ht == NULL) {
		printf("hash_create failed\n");
		exit(1);
	}

	/* First phase: insert some data. */
	printf("\nInsert phase:\n");
	for (int i = 0; i < N * 2; i++) {
		k = (char*) malloc(sizeof(char) * BUFLEN);
		snprintf(k, BUFLEN, "String %d", i % N);
		v = (long int*) malloc(sizeof(long int));
		*v = i;
		removed_key = NULL;
		removed_value = NULL;

		// The hash map takes ownership of the key and value:
		if (!hash_insert(ht, k, v, (void**) &removed_key, (void**) &removed_value)) {
			printf("Insertion of (%s, %ld) failed\n", k, *v);
		} else if (removed_value != NULL) {
			printf("Replaced (%s, %ld) while inserting (%s, %ld)\n",
					removed_key, *removed_value, k, *v);
			free(removed_key);
			free(removed_value);
		} else {
			printf("Inserted (%s, %ld)\n", k, *v);
		}
	}

	/* Second phase: look up some data. */
	printf("\nLookup phase:\n");
	for (int i = N - 1; i >= 0; i--) {
		snprintf(strbuf, BUFLEN, "String %d", i);
		if (!hash_lookup(ht, strbuf, (void**) &v)) {
			printf("Entry for %s not found\n", strbuf);
		} else {
			printf("%s -> %ld\n", strbuf, *v);
		}
	}

	/* Look up a key that hasn't been inserted: */
	if (!hash_lookup(ht, not_found_key, (void**) &v)) {
		printf("Lookup of \"%s\" failed (as expected)\n", not_found_key);
	} else {
		printf("%s -> %ld (unexpected!)\n", not_found_key, *v);
	}

	/* Destroy the hash table and free things that we've allocated. Because
	 * we allocated both the keys and the values, we instruct the hash map
	 * to free both.
	 */
	hash_destroy(ht, true, true);

	return 0;
}
